/**
 *
 * Codice sorgente per configurare una centralina di monitoraggio della qualità
 * dell'aria ispirata a quelle del progetto ScienziAria di Comunità Circolare.
 *
 *   GUIDA PASSO PASSO per costruire la tua CENTRALINA FAI DA TE
 *   👉️ https://comunitacircolare.it/articoli/scienziaria-guida-centralina
 *
 * Questa versione di base è pensata espressamente per chi non ha particolari
 * competenze tecniche ed è incentrata quindi sulla semplicità di realizzazione.
 *
 * Se non hai mai programmato, NIENTE PANICO! Ti basta inserire i tre parametri
 * richiesti qui sotto all'interno delle virgolette, senza toccare nient'altro.
 *
 * Altrimenti, modifica pure il codice come ti pare (a tuo rischio e pericolo).
 *
 * \author Michele Nuzzolese
 * \copyright Pubblico dominio con licenza Unlicense: https://unlicense.org
 * \sa Progetto ScienziAria: https://comunitacircolare.it/progetti/scienziaria
 *
 */

/* ## PARAMETRI RICHIESTI ## */

// 1. Nome della rete Wi-Fi, da ricopiare esattamente come appare nella lista di
// reti disponibili sul tuo smartphone o computer (includendo eventuali spazi o
// punteggiatura e rispettando la differenza tra lettere maiuscole e minuscole).
const char *NOME_RETE_WIFI = "scrivi qui il nome della tua rete Wi-Fi";

// 2. Password del Wi-Fi: occhio a maiuscole/minuscole/errori di battitura/etc.
const char *PASSWORD_WIFI = "qui dentro ci va la password";

// 3. `Write API Key` (la trovi nella tab `API Keys` del tuo canale ThingSpeak).
const char *WRITE_API_KEY = "0RM414VR41C4P1T0";

/* ## FINE CONFIGURAZIONI OBBLIGATORIE ##
 *
 * Congratulazioni! Non ti resta che caricare questo codice nella tua centralina
 * con Arduino IDE (per istruzioni passo passo c'è sempre la guida in alto 👆️).
 *
 * ## INIZIO PROGRAMMA (E PARAMETRI OPZIONALI) ##
 *
 * Il codice sorgente ti fa paura? Scappa ora, prima che sia troppo tardi! 😜️
 *
 * Nel seguito troverai tutti i parametri opzionali legati alla logica interna
 * della centralina. C'è qualche commento per facilitare la comprensione, ma non
 * potevo dilungarmi troppo. Nel dubbio, puoi sempre usare i valori predefiniti.
 *
 * Se invece vuoi spiegazioni più dettagliate, ecco l'articolo che fa per te:
 * https://comunitacircolare.it/articoli/scienziaria-approfondimenti-centraline
 *
 */

#include <WiFi.h>
#include <HTTPClient.h>
#include <NetworkClientSecure.h>
#include <esp32-hal-cpu.h>
#include <esp_sleep.h>

// I valori indicati di seguito si basano sulle istruzioni della guida e devono
// rispecchiare l'effettivo collegamento tra il sensore e la scheda ESP32:
const uint8_t PIN_RX2 = 26;  // Da collegare al pin TXD del sensore.
const uint8_t PIN_TX2 = 27;  // Da collegare al pin RXD del sensore.

// Frequenza con cui la centralina invia nuove rilevazioni (in minuti).
const uint8_t INTERVALLO_AGGIORNAMENTI_M = 4;

// Tempo di attesa per letture stabili dall'avvio della ventola (in secondi).
const uint8_t STABILIZZAZIONE_SENSORE_S = 30;
// Minimo intervallo di tempo tra due letture contigue (in millisecondi).
const uint32_t POLLING_SENSORE_MS = 1200;
// Numero massimo di rilevazioni del sensore su cui calcolare la media.
const uint8_t MAX_LETTURE_SENSORE = 6;

// Limite di tempo per tentare di connettersi al Wi-Fi (in millisecondi).
const uint32_t TIMEOUT_WIFI_MS = 10000;

// Certificato del server di ThingSpeak per la connessione sicura via HTTPS.
const char *CA_CERT =
  "-----BEGIN CERTIFICATE-----\n"
  "MIIDjjCCAnagAwIBAgIQAzrx5qcRqaC7KGSxHQn65TANBgkqhkiG9w0BAQsFADBh\n"
  "MQswCQYDVQQGEwJVUzEVMBMGA1UEChMMRGlnaUNlcnQgSW5jMRkwFwYDVQQLExB3\n"
  "d3cuZGlnaWNlcnQuY29tMSAwHgYDVQQDExdEaWdpQ2VydCBHbG9iYWwgUm9vdCBH\n"
  "MjAeFw0xMzA4MDExMjAwMDBaFw0zODAxMTUxMjAwMDBaMGExCzAJBgNVBAYTAlVT\n"
  "MRUwEwYDVQQKEwxEaWdpQ2VydCBJbmMxGTAXBgNVBAsTEHd3dy5kaWdpY2VydC5j\n"
  "b20xIDAeBgNVBAMTF0RpZ2lDZXJ0IEdsb2JhbCBSb290IEcyMIIBIjANBgkqhkiG\n"
  "9w0BAQEFAAOCAQ8AMIIBCgKCAQEAuzfNNNx7a8myaJCtSnX/RrohCgiN9RlUyfuI\n"
  "2/Ou8jqJkTx65qsGGmvPrC3oXgkkRLpimn7Wo6h+4FR1IAWsULecYxpsMNzaHxmx\n"
  "1x7e/dfgy5SDN67sH0NO3Xss0r0upS/kqbitOtSZpLYl6ZtrAGCSYP9PIUkY92eQ\n"
  "q2EGnI/yuum06ZIya7XzV+hdG82MHauVBJVJ8zUtluNJbd134/tJS7SsVQepj5Wz\n"
  "tCO7TG1F8PapspUwtP1MVYwnSlcUfIKdzXOS0xZKBgyMUNGPHgm+F6HmIcr9g+UQ\n"
  "vIOlCsRnKPZzFBQ9RnbDhxSJITRNrw9FDKZJobq7nMWxM4MphQIDAQABo0IwQDAP\n"
  "BgNVHRMBAf8EBTADAQH/MA4GA1UdDwEB/wQEAwIBhjAdBgNVHQ4EFgQUTiJUIBiV\n"
  "5uNu5g/6+rkS7QYXjzkwDQYJKoZIhvcNAQELBQADggEBAGBnKJRvDkhj6zHd6mcY\n"
  "1Yl9PMWLSn/pvtsrF9+wX3N3KjITOYFnQoQj8kVnNeyIv/iPsGEMNKSuIEyExtv4\n"
  "NeF22d+mQrvHRAiGfzZ0JFrabA0UWTW98kndth/Jsw1HKj2ZL7tcu7XUIOGZX1NG\n"
  "Fdtom/DzMNU+MeKNhJ7jitralj41E6Vf8PlwUHBHQRFXGU7Aj64GxJUTFy8bJZ91\n"
  "8rGOmaFvE7FBcf6IKshPECBV1/MUReXgRPTqh5Uykw7+U0b6LJ3/iyK5S9kJRaTe\n"
  "pLiaWN0bfVKfjllDiIGknibVb63dDcY3fe0Dkhvld1927jyNxF1WW6LZZm6zNTfl\n"
  "MrY=\n"
  "-----END CERTIFICATE-----";
// Limite di tempo per effettuare l'handshake SSL (in secondi).
const uint32_t TIMEOUT_HANDSHAKE_SSL_S = 5;

// Limite di tempo per stabilire la connessione con il server (in millisecondi).
const uint32_t TIMEOUT_CONNESSIONE_CLIENT_MS = 5000;
// Limite di tempo per ricevere la risposta del server (in millisecondi).
const uint32_t TIMEOUT_RISPOSTA_SERVER_MS = 5000;
// Numero massimo di tentativi di invio della richiesta GET.
const uint32_t MAX_TENTATIVI_RICHIESTA = 3;
// Fattore di backoff per nuovi tentativi di invio (in millisecondi).
const uint32_t FATTORE_BACKOFF_MS = 600;
// Limite massimo di attesa prima di un nuovo tentativo (in millisecondi).
const uint32_t MAX_BACKOFF_MS = 3000;

// Per semplicità, usiamo delle variabili globali per le polveri sottili:
uint16_t pm1 = 0;   // Concentrazione di PM1 rilevata dal sensore.
uint16_t pm25 = 0;  // Concentrazione di PM2,5 rilevata dal sensore.
uint16_t pm10 = 0;  // Concentrazione di PM10 rilevata dal sensore.

/* ## FINE PARAMETRI OPZIONALI ##
 *
 * Da qui in poi ci sono le funzioni: "pezzi di codice" con uno scopo specifico,
 * che possono essere richiamate all'occorrenza dal programma principale.
 *
 * A proposito, uno sketch inizia sempre con la funzione `setup()` (eseguita una
 * sola volta) per poi passare alla funzione `loop()` (ripetuta ad libitum).
 *
 * Nel nostro caso il `loop()` non serve, perché l'ESP32 riparte dal `setup()`
 * dopo che la centralina va in risparmio energetico (v. modalità `deep sleep`).
 *
 * ## INIZIO FUNZIONI HELPER ##
 *
 * Come suggerisce il nome, sono funzioni scritte per aiutare con determinati
 * compiti (nella fattispecie, leggere il sensore e ibernare la centralina).
 *
 */

/**
 * Inizializza la comunicazione con il sensore e lo predispone alla lettura.
 *
 * Nota: abbiamo optato per un'implementazione meno efficiente ma più semplice e
 * robusta, per prevenire potenziali rilevazioni fallite in casi limite (sulla
 * base del funzionamento del sensore). Per approfondimenti, vedi l'articolo:
 * https://comunitacircolare.it/articoli/scienziaria-approfondimenti-centraline
 *
 */
void initPM() {

  // Inizializziamo la comunicazione seriale
  Serial2.begin(9600, SERIAL_8N1, PIN_RX2, PIN_TX2);

  // Comando per mettere il sensore in `passive mode` (invio dati su richiesta)
  uint8_t comandoPassiveMode[7] = { 0x42, 0x4D, 0xE1, 0x00, 0x00, 0x01, 0x70 };
  // Inviamo il comando (a cui il sensore risponde con un messaggio di conferma)
  Serial2.write(comandoPassiveMode, 7);

  // Assicuriamoci che il sensore sia pronto per la lettura prima di proseguire
  delay(942);  // tempo di attesa = test empirici + ampio margine di sicurezza
}

/**
 * Legge il sensore (presupponendo che la comunicazione sia già stata aperta) e
 * memorizza le misurazioni di polveri sottili nelle apposite variabili globali.
 *
 * @return `true` se la lettura ha esito positivo, `false` altrimenti.
 */
bool readPM() {

  // Scartiamo eventuali dati spuri che potrebbero interferire con la lettura
  while (Serial2.available()) {
    Serial2.read();
  }

  // Comando per richiedere una lettura quando il sensore è in `passive mode`
  uint8_t comandoRead[7] = { 0x42, 0x4D, 0xE2, 0x00, 0x00, 0x01, 0x71 };
  // Inviamo il comando (a cui il sensore risponde con i dati della lettura)
  Serial2.write(comandoRead, 7);
  // Aspettiamo che la trasmessione seriale del comando sia avvenuta per intero
  Serial2.flush();
  // Aspettiamo la risposta (dovrebbero bastare 34 ms, ma noi siamo prudenti :P)
  delay(42);

  // Assicuriamoci di aver ricevuto tutti e soli i 32 byte che ci aspettiamo in
  // base al protocollo di trasporto (rif. scheda tecnica del sensore PMS5003)
  if (Serial2.available() != 32) {
    return false;  // risposta assente o non conforme al protocollo di trasporto
  }

  uint8_t datiSensore[32] = { 0 };
  uint16_t checksumCalcolato = 0;

  // Prepariamo i dati da elaborare e i checksum per i controlli di integrità
  for (uint8_t i = 0; i < 32; i++) {
    datiSensore[i] = Serial2.read();
    checksumCalcolato += datiSensore[i];
  }
  checksumCalcolato -= datiSensore[30] + datiSensore[31];

  // Validiamo i dati ricevuti e controlliamo l'esito della trasmissione seriale
  if (checksumCalcolato == 0 || checksumCalcolato == 0xAB) {
    return false;  // sensore non disponibile (tutti i dati uguali a 0)
  }
  uint16_t checksumRicevuto = (datiSensore[30] << 8) | datiSensore[31];
  if (checksumRicevuto != checksumCalcolato) {
    return false;  // errore di trasmissione (checksum non corrispondenti)
  }

  // Elaboriamo le rilevazioni di PM e memorizziamole nelle rispettive variabili
  pm1 = (datiSensore[10] << 8) | datiSensore[11];
  pm25 = (datiSensore[12] << 8) | datiSensore[13];
  pm10 = (datiSensore[14] << 8) | datiSensore[15];

  return true;  // lettura effettuata con successo
}

/**
 * Mette la centralina in ibernazione per risparmiare energia quando non rileva.
 *
 * Nota: in questa versione di base, solo l'ESP32 va in modalità sleep. Quindi:
 *
 *  1. La ventola del sensore PM gira sempre (e consuma corrente).
 *  2. Al risveglio dell'ESP32 sono già disponibili rilevazioni stabili.
 *  3. L'attività continuativa potrebbe influire sul tempo di vita del sensore.
 *
 * (Anche il sensore ha una modalità sleep, ma qui non viene usata per mantenere
 * più semplice la fase di assemblaggio. Per una variante più efficiente, vedi:
 * https://comunitacircolare.it/articoli/scienziaria-approfondimenti-centraline)
 */
void hibernate() {

  // Disabilitiamo il Wi-Fi (v. documentazione "Sleep Modes" dell'ESP32)
  WiFi.mode(WIFI_OFF);

  // Disattiviamo esplicitamente i domini che non ci servono (attivandoli prima
  // a causa di un bug, v. https://esp32.com/viewtopic.php?p=132473#p132473)
  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_PERIPH, ESP_PD_OPTION_ON);
  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_PERIPH, ESP_PD_OPTION_OFF);
  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_SLOW_MEM, ESP_PD_OPTION_ON);
  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_SLOW_MEM, ESP_PD_OPTION_OFF);
  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_FAST_MEM, ESP_PD_OPTION_ON);
  esp_sleep_pd_config(ESP_PD_DOMAIN_RTC_FAST_MEM, ESP_PD_OPTION_OFF);
  esp_sleep_pd_config(ESP_PD_DOMAIN_XTAL, ESP_PD_OPTION_ON);
  esp_sleep_pd_config(ESP_PD_DOMAIN_XTAL, ESP_PD_OPTION_OFF);

  // Calcoliamo il tempo mancante fino al prossimo ciclo in millisecondi
  uint32_t sleepMillis = 60000UL * INTERVALLO_AGGIORNAMENTI_M - millis();
  esp_deep_sleep(1000ULL * sleepMillis);  // e lo convertiamo in microsecondi

  // Nota: in realtà l'ESP32 non tiene il tempo con gran precisione (a lungo
  // andare tende a perdere qualche secondo), ma ai nostri fini va bene così.
}

/* ## FINE FUNZIONI HELPER (E INIZIO PROGRAMMA PRINCIPALE) ## */

/**
 * Qui dentro c'è un intero ciclo di misurazioni e invio dei dati rilevati.
 *
 * Questa funzione infatti viene eseguita non solo appena la centralina riceve
 * corrente, ma anche ogni volta che si risveglia dalla modalità ibernazione.
 */
void setup() {

  /*** IMPOSTAZIONI PRELIMINARI ***/

  // Se siamo al primo ciclo della centralina (e cioè al primo avvio dell'ESP32)
  if (esp_sleep_get_wakeup_cause() != ESP_SLEEP_WAKEUP_TIMER) {
    // Aspettiamo che il sensore si stabilizzi prima di richiedere dati
    esp_deep_sleep(1000000ULL * STABILIZZAZIONE_SENSORE_S);
  }
  // Nota: la ventola del sensore di particelle ha bisogno di tempo per creare
  // un flusso d'aria costante (indispensabile per misurazioni affidabili), ma
  // finché è alimentata poi continua a girare anche se l'ESP32 è in deep sleep
  // (v. anche il commento sulla funzione `hibernate()` poco sopra). Per questo
  // basta stabilizzare il sensore una sola volta, ovvero quando centralina è
  // appena stata collegata alla corrente (e la ventola parte da ferma).

  // Assicuriamoci che la frequenza della CPU sia 80 MHz per risparmiare energia
  if (getCpuFrequencyMhz() != 80) {
    setCpuFrequencyMhz(80);
  }
  // Nota: se la imposti con Arduino IDE prima di caricare lo sketch (dal menu
  // `Tools` o `Strumenti` -> `CPU Frequency`) potrai fare a meno sia delle tre
  // righe di codice precedenti che della libreria `esp32-hal-cpu.h` in alto. E
  // già che ci sei, puoi anche controllare che `Core Debug Level` sia impostato
  // su `None` (tanto gli eventuali log non si potrebbero comunque leggere, dato
  // che la centralina non sarà collegata al PC). Queste voci di menu però sono
  // disponibili solo se hai già selezionato una scheda ESP32 compatibile.

  /*** LETTURA DEL SENSORE ***/

  // Inizializziamo la comunicazione con il sensore
  initPM();

  // Totale di letture effettuate in questo ciclo di esecuzione della centralina
  uint8_t lettureRiuscite = 0, lettureFallite = 0;
  // Media delle misurazioni effettuate in questo ciclo
  float pm10media = 0, pm25media = 0, pm1media = 0;

  // Leggiamo il sensore ad intervalli definiti in `POLLING_SENSORE_MS` per il
  // numero di volte definito in `MAX_LETTURE_SENSORE` e ne calcoliamo la media
  while (1) {  // usiamo `break` per non aspettare a vuoto dopo l'ultima lettura

    // Contiamo sia i tentativi di lettura falliti che quelli riusciti
    if (!readPM()) {
      lettureFallite++;
    } else {
      lettureRiuscite++;

      // Calcoliamo la media cumulativa delle misurazioni effettuate
      pm10media = (pm10 + (lettureRiuscite - 1) * pm10media) / lettureRiuscite;
      pm25media = (pm25 + (lettureRiuscite - 1) * pm25media) / lettureRiuscite;
      pm1media = (pm1 + (lettureRiuscite - 1) * pm1media) / lettureRiuscite;
    }

    // Se abbiamo raggiunto il limite di letture, usciamo dal `while`
    if (lettureRiuscite + lettureFallite >= MAX_LETTURE_SENSORE)
      break;

    // Altrimenti attendiamo fino alla prossima rilevazione
    delay(POLLING_SENSORE_MS);
  }
  // Nota: chiudere esplicitamente la comunicazione con il sensore è superfluo,
  // perché tutta la memoria dell'ESP32 verrà comunque deallocata a fine ciclo.

  // Percentuale di successo del sensore durante questo ciclo della centralina
  float esitoLettureSensore = 100.0 * lettureRiuscite / (lettureRiuscite + lettureFallite);

  /*** CONNESSIONE AL WI-FI ***/

  uint32_t inizioWifiMillis = millis();
  WiFi.begin(NOME_RETE_WIFI, PASSWORD_WIFI);

  // Controlliamo il Wi-Fi finché siamo connessi o sforiamo il limite di tempo
  while (WiFi.status() != WL_CONNECTED) {
    if ((millis() - inizioWifiMillis) > TIMEOUT_WIFI_MS) {
      // Senza Wi-Fi non possiamo inviare i dati, quindi mettiamo la centralina
      // in modalità risparmio energetico per poi riprovarci al prossimo ciclo.
      hibernate();
    }
    // Nota: il timeout è necessario per via di un bug che potrebbe verificarsi
    // in condizioni di bassa potenza del segnale Wi-Fi, in cui l'ESP32 smette
    // di tentare la riconnessione e resta indefinitamente bloccata in questo
    // `while` (anche qualora la rete dovesse tornare a essere disponibile).
    //
    // A prescindere dall'implementazione dell'ESP32, nel nostro caso prevedere
    // un timeout è comunque positivo in termini di efficienza energetica (come
    // misura preventiva verso eventuali indisponibilità prolungate del Wi-Fi).
  }

  // Potenza del segnale Wi-Fi (Received Signal Strength Indicator)
  int8_t rssi = WiFi.RSSI();
  // Tempo di connessione al Wi-Fi (in secondi)
  float tempoWifi = (millis() - inizioWifiMillis) / 1000.0;

  /*** INVIO DATI A THINGSPEAK ***/

  // Creiamo la richiesta nel formato definito dall'API di ThingSpeak
  String uri = "/update?api_key=" + String(WRITE_API_KEY);
  // Inviamo i dati sulle polveri sottili solo se è riuscita almeno una lettura
  if (lettureRiuscite > 0) {
    uri += "&field1=" + String(pm10media, 2)
           + "&field2=" + String(pm25media, 2)
           + "&field3=" + String(pm1media, 2);
  }
  // Invece questi dati di controllo li inviamo sempre e comunque
  uri += "&field4=" + String(esitoLettureSensore, 2)
         + "&field5=" + String(rssi)
         + "&field6=" + String(tempoWifi, 3);

  // Prepariamo i client per la richiesta GET via HTTPS
  NetworkClientSecure client;  // Client per la gestione del protocollo HTTPS
  HTTPClient httpClient;       // Client per la gestione della richiesta GET

  // Impostiamo il certificato e il timeout per l'handshake SSL/TLS
  client.setCACert(CA_CERT);
  client.setHandshakeTimeout(TIMEOUT_HANDSHAKE_SSL_S);

  // Impostiamo i timeout per la connessione del client e la risposta del server
  // (v. https://github.com/espressif/arduino-esp32/issues/1433#issuecomment-475875579)
  httpClient.setConnectTimeout(TIMEOUT_CONNESSIONE_CLIENT_MS);
  httpClient.setTimeout(TIMEOUT_RISPOSTA_SERVER_MS);

  // Nota: di default la connessione TCP viene lasciata aperta per il riuso, ma
  // purtroppo non sembra sopravvivere al riavvio dell'ESP32 nemmeno se i client
  // sono preservati nella memoria RTC. Il guadagno sarebbe comunque marginale.

  httpClient.begin(client, "api.thingspeak.com", 443, uri, true);

  uint8_t contatoreRichieste = 0;
  uint32_t attesaBackoff = 0;

  // Tentiamo l'invio ad intervalli calcolati usando un algoritmo di backoff
  // esponenziale in base ai parametri `FATTORE_BACKOFF_MS` e `MAX_BACKOFF_MS`
  // per un massimo numero di volte definito in `MAX_TENTATIVI_RICHIESTA`
  while (httpClient.GET() != HTTP_CODE_OK) {

    // Se abbiamo raggiunto il limite di tentativi, usciamo dal `while`
    if (++contatoreRichieste >= MAX_TENTATIVI_RICHIESTA)
      break;

    // Altrimenti attendiamo fino alla prossima richiesta
    attesaBackoff = FATTORE_BACKOFF_MS << (contatoreRichieste - 1);
    attesaBackoff = min(attesaBackoff, MAX_BACKOFF_MS);
    delay(attesaBackoff);
    // Nota: oggettivamente, non era affatto necessario un meccanismo di backoff
    // esponenziale: sarebbe bastato anche solo riprovare a intervalli fissi...
    // ma ormai è stato già implementato, quindi tanto vale lasciarlo lì! :P
  }
  // Nota: questo ciclo è finito, evitiamo di deallocare i client per lo stesso
  // motivo per cui non abbiamo deinizializzato la comunicazione con il sensore.

  // Andiamo in modalità risparmio energetico fino al prossimo ciclo
  hibernate();
}

/**
 * Questa funzione è obbligatoria. Ma anche no ¯\_(ツ)_/¯
 */
void loop() {}
